//
// File :    LCD.ccc
// Purpose : provides function to send data to lcd
//

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <syslog.h>
#include <malloc.h>
#include <termios.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <poll.h>

#include "LCD.hh"

namespace LCD
{

/////////////////////////////////////////////////////////
//
// defines
//
/////////////////////////////////////////////////////////

// maximal lenght of LCD buffer
#define MAX_LEN  32

// access framebuffer and change-map
#define LCD_FB( x, y )  lcd_fb[ (y*lcd_w) + x ]
#define LCD_CM( x, y )  lcd_chg_map[ (y*lcd_w) + x ]

// LED settings
#define LED_GREEN  0
#define LED_RED    1

// minimum, maximum
#define MIN( x, y ) ((x) < (y) ? (x) : (y))
#define MAX( x, y ) ((x) > (y) ? (x) : (y))

/////////////////////////////////////////////////////////
//
// variables
//
/////////////////////////////////////////////////////////

// framebuffer for the LCD
static char * lcd_fb = NULL;

// change-map of the framebuffer
static char * lcd_chg_map;

// dimension of the LCD
static int    lcd_w = 0;
static int    lcd_h = 0;

// file descriptor of the tty
static int    lcd_tty_fd;

static bool   use_green_led;
static bool   use_red_led;
static int    verbose;

// pos. of the cursor
static int    cur_x = 0, cur_y = 0;

/////////////////////////////////////////////////////////
//
// functions
//
/////////////////////////////////////////////////////////

//
// initialise LCD functions
//
void
init ( int w, int h, const char * device )
{
    struct termios portset;

    //
    // open IO port and set it up
    //
    
    if ( (lcd_tty_fd = open( device, O_RDWR | O_NOCTTY )) < 0 )
    {
        printf( "(LCD) main : error while opening \"%s\" : %s\n",
                device, strerror( errno ) );
        return;
    }

    tcgetattr( lcd_tty_fd, & portset );
    cfmakeraw( & portset );
    cfsetospeed( & portset, B9600 );
    tcsetattr( lcd_tty_fd, TCSANOW, & portset );

    //
    // set up framebuffer and LEDs
    //
    
    lcd_w = w;
    lcd_h = h;

    lcd_fb      = new char[ w*h ];
    lcd_chg_map = new char[ w*h ];

    for ( int j = 0; j < h; j++ )
    {
        for ( int i = 0; i < w; i++ )
        {
            LCD_FB( i, j ) = ' ';
            LCD_CM( i, j ) = 1;
        }
    }

    no_cursor();

    use_green_led = use_red_led = true;
    
    if ( use_green_led ) led_on( LED_GREEN );
    else                 led_off( LED_GREEN );

    if ( use_red_led ) led_on( LED_RED );
    else               led_off( LED_RED );
}

//
// gather lcd status
//
int
width ()
{
    return lcd_w;
}

int
height ()
{
    return lcd_h;
}

//
// write message to LCD
//
void
write ( const char * msg )
{
    int  len;

    if ( msg == NULL )
        return;

    len = strlen( msg );
    
    if ( len > MAX_LEN-2 )
    {
        printf( "(LCD) write : message too long; truncated\n" );
        return;
    }

    if ( verbose )
    {
        char buf2[64];
        int  n = len;

        strncpy( buf2, msg, sizeof(buf2) );

        // just for pretty: replay \n by "\n"
        if ( buf2[ n - 1 ] == '\n' )
        {
            buf2[n-1] = '\\';
            buf2[n]   = 'n';
            buf2[n+1] = '\0';
        }
        
        printf( "(LCD) write : message = [%s], %d\n", buf2, n );
    }
    
    if ( ::write( lcd_tty_fd, msg, len ) < 0 )
        printf( "(LCD) write : error while writing to tty : %s\n",
                strerror( errno ) );
}

//
// read message from LCD with given timeout in millisec.
// - return size of message or -1 if an error occured
//
int
read ( char * buf, int len, int timeout )
{
    //
    // wait idletime and check if something is send from LCD
    //
    
    struct pollfd   pfd;
    
    pfd.fd     = lcd_tty_fd;
    pfd.events = POLLIN | POLLPRI;
    
    int n = poll( & pfd, 1, timeout );
    
    if ( n == -1 )
        printf( "(main) error while polling tty : %s\n",
                strerror( errno ) );
    else if (( n > 0 ) && (pfd.revents & POLLIN))
    {
        if ((n = ::read( lcd_tty_fd, buf, len )) < 0 )
        {
            printf( "(read_tty) error while reading from tty : %s\n",
                    strerror( errno ) );
        }
    }

    return n;
}

//
// clear LCD
//
void
clear ()
{
    for ( int j = 0; j < lcd_h; j++ )
    {
        for ( int i = 0; i < lcd_w; i++ )
            putc_xy( i, j, ' ' );
    }            
}

//
// write message to LCD at position (x,y)
//
void
puts_xy ( int x, int y, const char * msg )
{
    int   n;
    
    if ( msg == NULL )
        return;

    n = strlen( msg );
    
    for ( int i = 0; i < n; i++ )
        putc_xy( x+i, y, msg[i] );
}

//
// write character to LCD at position (x,y)
//
void
putc_xy ( int x, int y, unsigned char c )
{
    if (( x >= 0 ) && ( x < lcd_w ) &&
        ( y >= 0 ) && ( y < lcd_h ))
    {
        if ( LCD_FB( x, y ) != c )
        {
            LCD_CM( x, y ) = 1;
            LCD_FB( x, y ) = c;
        }
    }
    else
        printf( "(LCD) putc_xy : writing outside LCD\n" );
}

//
// flush framebuffer to LCD
//
void
flush ()
{
    bool  transfer = false;
    char  cmd[4] = "d=\0";
    char  buf[3] = " \0";
    int   old_x  = cur_x;
    int   old_y  = cur_y;
    
    //
    // check each byte in framebuffer and only write changed
    // positions
    //

    for ( int j = 0; j < lcd_h; j++ )
    {
        for ( int i = 0; i < lcd_w; i++ )
        {
            if ( LCD_CM( i, j ) != 0 )
            {
                if ( ! transfer )
                {
                    // init transfer by setting cursor
                    set_cursor( i, j );

                    // and by starting the command string
                    write( cmd );
                    
                    transfer = true;
                }

                buf[0] = LCD_FB( i, j );
                write( buf );

                LCD_CM( i, j ) = 0;
            }
            else if ( transfer )
            {
                // finish transfer
                write( "\n" );
                transfer = false;
            }
        }

        // finish transfer at end of line
        if ( transfer )
        {
            write( "\n" );
            transfer = false;
        }
    }

    set_cursor( old_x, old_y );
}

//
// update all positions
//
void
invalid ()
{
    for ( int j = 0; j < lcd_h; j++ )
    {
        for ( int i = 0; i < lcd_w; i++ )
            LCD_CM( i, j ) = 1;
    }
}

//
// control cursor of LCD
//

void
set_cursor ( int x, int y )
{
    char  buf[16];

    snprintf( buf, sizeof(buf), "g=%02d%02d\n", x, y );
    write( buf );

    cur_x = x;
    cur_y = y;
}

void
no_cursor ()
{
    write( "c=i\n" );
}

void
normal_cursor ()
{
    write( "c=n\n" );
}

void
blink_cursor ()
{
    write( "c=b\n" );
}

//
// control LEDs of LCD
//

void
led_on  ( int l )
{
    if      ( l == LED_GREEN ) write( "l=g1\n" );
    else if ( l == LED_RED   ) write( "l=r1\n" );
}

void
led_off ( int l )
{
    if      ( l == LED_GREEN ) write( "l=g0\n" );
    else if ( l == LED_RED   ) write( "l=r0\n" );
}

//
// programable characters
//
void
set_hbars ( void )
{
    write( "p=h\n" );
}

void
set_vbars ( void )
{
    write( "p=v\n" );
}

}// namespace
