//
// Project : AVR LCD
// File    : Menu.cc
// Purpose : class for a menu-widget
//

#include <string.h>

#include "LCD.hh"
#include "Application.hh"
#include "Menu.hh"

namespace LWL
{

////////////////////////////////////////
//
// defines
//
////////////////////////////////////////

#define IS_MENU( m )  (dynamic_cast< const Menu * >( m ) != NULL)

#ifndef MIN
#define MIN( x, y )  ((x) < (y) ? (x) : (y))
#endif

////////////////////////////////////////
//
// Menu::Entry
//
////////////////////////////////////////

Menu::Entry::Entry ( const char * name, int cmd )
{
    strncpy( _name, name, sizeof(_name) );
    _cmd = cmd;
    _parent = NULL;
}

Menu::Entry::~Entry ()
{
    while ( _submenus.size() > 0 )
        delete _submenus.behead();
}

//
// manage submenus
//

Menu::Entry *
Menu::Entry::add_submenu ( Menu::Entry * submenu )
{
    if ( submenu != NULL )
    {
        _submenus.append( submenu );
        submenu->set_parent( this );
    }// if
    
    return this;
}

////////////////////////////////////////
//
// Menu
//
////////////////////////////////////////

//
// constructor and destructor
//

Menu::Menu ( int x, int y, int w, int h, Window * parent )
        : Window( x, y, w, h, parent )
{
    _pos = 0;

    _lb = _ub = 0;

    _act_submenu = _menu = new Entry( "Main" );
}

Menu::~Menu ()
{
    delete _menu;
}

//
// event handling
//

void
Menu::handle_event ( Event & e )
{
    if ( e.type() == Event::EV_KEYBOARD )
    {
        switch ( e.keycode() )
        {
        case BTN_DOWN :    // go up in list
            if ( _pos < _act_submenu->submenus().size()-1 )
                _pos++;
            e.clear();
            break;
            
        case BTN_UP   :    // go up in list
            if ( _pos > 0 )
                _pos--;
            e.clear();
            break;
            
        case BTN_MIDDLE :  // go down in menu hierarchy
            select_active();
            e.clear();
            break;
                
        case BTN_LEFT :    // go up in menu hierarchy
            if (( _act_submenu != NULL ) && ( _act_submenu->parent() != NULL ))
            {
                set_active( _act_submenu->parent() );
                e.clear();
            }// if
            break;
        }// switch

        if ( _pos >= _ub )
        {
            _ub++;
            _lb++;
        }// if
        
        if ( _pos < _lb )
        {
            _ub--;
            _lb--;
        }// if
    }// else
}

//
// draw window content
//

void
Menu::draw ()
{
    fill( ' ' );

    print_str( 2, 0, _act_submenu->name() );
    
    if ( _act_submenu->parent() != NULL )
        print_char( 1, 0, CHAR_ARROW_LEFT );
    
    unsigned  i = 1, p = 0;
    SLL< Entry * >::Iterator iter = _act_submenu->submenus().first();
    
    while (( p < _ub ) && ! iter.eol() )
    {
        if (( _lb <= p ) && ( p < _ub ))
        {
            print_str( 3, i, iter()->name() );
            
            if ( p == _pos )
                print_char( 2, i, CHAR_ARROW_RIGHT );
            
            i++;
        }// if
        
        ++p;
        ++iter;
    }// while
}
    
//
// manage submenus
//

Menu *
Menu::add_entry ( Entry * entry )
{
    if (( _menu != NULL ) && ( entry != NULL ))
    {
        _menu->add_submenu( entry );

        _ub++;
        _ub = MIN( _ub, height()-1 );
    }// if
    
    return this;
}

void
Menu::select_active ()
{
    unsigned  i = 0;
    SLL< Entry * >::Iterator iter = _act_submenu->submenus().first();
    
    while ( ! iter.eol() )
    {
        if ( _pos == i )
        {
            if ( iter()->command() >= 0 )
            {
                // create event
                Event  ev;
                
                ev.set_command( iter()->command() );
                global_app->add_event( ev );
            }// if
            else
                set_active( iter() );
            
            break;
        }// if
        
        ++i;
        ++iter;
    }// while
}

void
Menu::set_active ( Menu::Entry * menu )
{
    _act_submenu = menu;

    _lb  = 0;
    _pos = 0;
    
    if ( menu != NULL )
        _ub  = MIN( menu->submenus().size(), height()-1 );
    else
        _ub = 0;
}

}// namespace
