#ifndef __SLL_HH
#define __SLL_HH

//
//  Project   : misc
//  File      : TSLL.hh
//  Author    : Ronald Kriemann
//  Purpose   : Template class for a single concatenated list
//

#include <assert.h>

namespace LWL
{

////////////////////////////////////////////////////
// 
// Template class list
//
template <class T> 
class SLL
{
public:
    ////////////////////////////////////////////////////
    //
    // class for a single item in the list
    //

    class Item
    {
    public:
        T       _data;
        Item  * _next;
        
    public:
        Item ( const T  & e, 
               Item    * n = NULL )
                : _data( e ), _next( n )
        {}
    };

    //////////////////////////////////////////////////////////
    // 
    // class for an iterator for the previous list
    //

    class Iterator
    {
        friend class SLL<T>;
        
    protected:
        // currently selected item
        Item  * _item;
        
    public:
        //
        // constructor and destructor
        //

        Iterator ()                        : _item(NULL) {}
        Iterator ( Item * item )           : _item(item) {}
        Iterator ( const Iterator & iter ) : _item(iter._item) {}
        
        // get status (end-of-list)
        bool eol () const { return (_item == NULL); }
        
        // increment
        Iterator & next () { return ++*this; }
        
        // access actual element
        T &       data ()       { assert(_item != NULL); return _item->_data; }
        const T & data () const { assert(_item != NULL); return _item->_data; }

        //
        // usual operators
        //

        // copy
        Iterator & operator = ( const Iterator & i ) { _item = i._item; return *this; }

        // access items
        T &        operator () ()       { assert(_item != NULL); return _item->_data; }
        const T &  operator () () const { assert(_item != NULL); return _item->_data; }

        // iterate (prefix/postfix)
        Iterator & operator ++ ()    { if (_item) _item = _item->_next; return *this; }
        Iterator   operator ++ (int) { Iterator tmp(*this); ++*this; return tmp; }

        // dereference
        T &        operator *  ()    { assert(_item != NULL); return   _item->_data; }
        T *        operator -> ()    { assert(_item != NULL); return & _item->_data; }

        // compare
        bool       operator == ( const Iterator & i ) const { return (_item == i._item); }
        bool       operator != ( const Iterator & i ) const { return (_item != i._item); }
    };

protected:
    // first and last element of list
    Item        * _first, * _last;
    
    // size of the list aka number of stored items
    unsigned int  _size;

public:
    /////////////////////////////////////////////////
    //
    // constructors and destructors
    //

    SLL ()                   : _first(NULL), _last(NULL), _size(0) {}
    SLL ( const SLL<T> & l ) : _first(NULL), _last(NULL), _size(0) { copy( l ); }

    ~SLL () { if (_size > 0) remove_all(); }


    /////////////////////////////////////////////////
    //
    // access member variables
    //

    unsigned int size () const { return _size; }

    Iterator       first ()       { return _first; }
    const Iterator first () const { return _first; }

    Iterator       last  ()       { return _last; }
    const Iterator last  () const { return _last; }


    // same as above but access real data
    T &       first_item ()       { assert(_first != NULL); return _first->_data; }
    const T & first_item () const { assert(_first != NULL); return _first->_data; }
    
    T &       last_item ()       { assert(_last != NULL); return _last->_data; }
    const T & last_item () const { assert(_last != NULL); return _last->_data; }
    
    /////////////////////////////////////////////////
    //
    // methods for manipulating list
    //

    // add a new object to the begin/end of list
    SLL<T> & prepend ( const T & elem );
    SLL<T> & append  ( const T & elem );

    // remove (next) element from list
    void remove ( const T & elem );
    void remove ( Iterator & i ) { assert(i._item != NULL); remove(i._item->_data); }
    void remove_first ();

    // return first item and remove it from list
    T behead ();
    
    // remove all elements from list
    void remove_all ();

    // copy given list into local
    void copy ( const SLL< T > & list );
    
    /////////////////////////////////////////////////
    //
    // search/sort related methods
    //

    // look if element is in list
    bool is_in ( const T & elem );

    /////////////////////////////////////////////////
    //
    // misc.
    //
    
    // return size in bytes used by this object
    unsigned long byte_size () const;
};

#include "SLL.cc"

}// namespace

#endif   // __SLL_HH
